import logging


class LoggingMixin:
    """
    Handy logging mixin that allows me to use a logger without having to
    declare it globally.
    """

    _logger = None

    @property
    def logger(self):

        if self._logger:
            return self._logger

        self._logger = logging.getLogger(
            ".".join(["pyletheia", __name__, self.__class__.__name__]))

        return self.logger



import os
from hashlib import md5
from unittest import mock

import piexif
from PIL import Image

from aletheia.file_types import JpegFile

from ..base import TestCase


class JpegTestCase(TestCase):

    def test_get_raw_data_from_path(self):
        unsigned = os.path.join(self.DATA, "test.jpg")
        self.assertEqual(
            md5(JpegFile(unsigned, "").get_raw_data()).hexdigest(),
            "cc96a1bff6c259f0534f191e83cfdf0e"
        )

    def test_get_raw_data_from_image(self):
        im = Image.open(os.path.join(self.DATA, "test.jpg"))
        self.assertEqual(
            md5(JpegFile(im, "").get_raw_data()).hexdigest(),
            "cc96a1bff6c259f0534f191e83cfdf0e"
        )

    def test_sign_from_path(self):

        path = self.copy_for_work("test.jpg")

        f = JpegFile(path, "")
        f.generate_signature = mock.Mock(return_value="signature")
        f.generate_payload = mock.Mock(return_value="payload")
        f.sign(None, None)

        exif = piexif.load(path)
        self.assertEqual(exif["0th"][piexif.ImageIFD.HostComputer], b"payload")

    def test_sign_from_image(self):

        path = self.copy_for_work("test.jpg")
        im = Image.open(path)

        f = JpegFile(im, "")
        f.generate_signature = mock.Mock(return_value="signature")
        f.generate_payload = mock.Mock(return_value="payload")
        signed_im = f.sign(None, None)

        exif = piexif.load(signed_im.info["exif"])
        self.assertEqual(exif["0th"][piexif.ImageIFD.HostComputer], b"payload")

    def test_verify_from_path_no_signature(self):

        path = self.copy_for_work("test.jpg")

        f = JpegFile(path, "")
        self.assertFalse(f.verify())

    def test_verify_from_path(self):

        path = self.copy_for_work("test-signed.jpg")

        f = JpegFile(path, "")
        f.verify_signature = mock.Mock(return_value=True)
        self.assertTrue(f.verify())

    def test_verify_from_image_no_signature(self):

        path = self.copy_for_work("test.jpg")
        im = Image.open(path)

        f = JpegFile(im, "")
        self.assertFalse(f.verify())

    def test_verify_from_image(self):

        path = self.copy_for_work("test-signed.jpg")
        im = Image.open(path)

        f = JpegFile(im, "")
        f.verify_signature = mock.Mock(return_value=True)
        self.assertTrue(f.verify())



import json
from io import BytesIO

import piexif
from PIL import Image
from PIL.JpegImagePlugin import JpegImageFile

from ..base import File


class JpegFile(File):

    SUPPORTED_TYPES = ("image/jpeg",)

    def get_raw_data(self):

        if isinstance(self.source, JpegImageFile):
            return self.source.tobytes()

        with Image.open(self.source) as im:
            return im.tobytes()

    def sign(self, private_key, public_key_url):
        """
        Use Pillow to capture the raw image data, generate a signature from it,
        and then use piexif to write said signature + where to find the public
        key to the image metadata in the following format:

          {"version": int, "public-key": url, "signature": signature}

        :param private_key     key  The private key used for signing
        :param public_key_url  str  The URL where you're storing the public key

        :return None if we're operating on a file. If we're working with an
                Image object, we return the signed object.
        """

        signature = self.generate_signature(private_key)

        self.logger.debug("Signature generated: %s", signature)

        payload = self.generate_payload(public_key_url, signature)

        if isinstance(self.source, JpegImageFile):
            return self._sign_inline(payload)

        exif = piexif.load(self.source)
        exif["0th"][piexif.ImageIFD.HostComputer] = payload
        piexif.insert(piexif.dump(exif), self.source)

    def _sign_inline(self, payload):
        """
        This isn't the most elegant way to do this, but I don't know of a
        better one.  The problem is that Pillow doesn't appear to have a means
        of editing EXIF data in-place, rather you have to write the stuff to
        the image object during the ``.save()`` step, so to work around this,
        we're just writing to memory and then re-opening it to return the
        newly-tagged object.

        Note that doing this subjects us to JPEG degradation as it's a lossy
        compression algorithm.  Better ideas are always welcome.

        :param payload: The JSON string we're writing to the image
        :return: A Pillow ``Image`` object with the new EXIF headers attached.
        """

        exif = {"0th": {}}
        if "exif" in self.source.info:
            exif = piexif.load(self.source.info.get("exif"))

        temp = BytesIO()
        exif["0th"][piexif.ImageIFD.HostComputer] = payload
        self.source.save(
            temp,
            "jpeg",
            quality="keep",
            subsampling="keep",
            qtables="keep",
            exif=piexif.dump(exif)
        )
        temp.seek(0)
        return Image.open(temp)

    def verify(self):
        """
        Attempt to verify the origin of an image by checking its local
        signature against the public key listed in the file.
        :return: boolean  ``True`` if verified, `False`` if not.
        """

        if isinstance(self.source, JpegImageFile):
            if "exif" not in self.source.info:
                self.logger.error("No EXIF data found in Image instance")
                return False
            exif = piexif.load(self.source.info["exif"])
        else:
            exif = piexif.load(self.source)

        try:
            data = json.loads(
                exif["0th"][piexif.ImageIFD.HostComputer].decode("utf-8"))
            key_url = data["public-key"]
            signature = data["signature"]
        except (KeyError, json.JSONDecodeError):
            self.logger.error("Invalid format, or no signature found")
            return False

        self.logger.debug("Signature found: %s", signature)

        return self.verify_signature(key_url, signature)



import os
from hashlib import md5
from unittest import mock

from mutagen.id3 import ID3

from aletheia.file_types import Mp3File

from ..base import TestCase


class Mp3TestCase(TestCase):

    def test_get_raw_data_from_path(self):
        unsigned = os.path.join(self.DATA, "test.mp3")
        self.assertEqual(
            md5(Mp3File(unsigned, "").get_raw_data().read()).hexdigest(),
            "8311d858697187b9cac8536221b5591d"
        )

    def test_sign_from_path(self):

        path = self.copy_for_work("test.mp3")

        f = Mp3File(path, "")
        f.generate_signature = mock.Mock(return_value="signature")
        f.generate_payload = mock.Mock(return_value="payload")
        f.sign(None, None)

        audio = ID3(path)
        self.assertEqual(audio["TPUB"], ["payload"])

    def test_verify_from_path_no_signature(self):

        path = self.copy_for_work("test.mp3")

        f = Mp3File(path, "")
        self.assertFalse(f.verify())

    def test_verify_from_path(self):

        path = self.copy_for_work("test-signed.mp3")

        f = Mp3File(path, "")
        f.verify_signature = mock.Mock(return_value=True)
        self.assertTrue(f.verify())



import binascii
import hashlib
import json
import os

import requests
import six

import magic
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, utils

from ..common import LoggingMixin
from ..exceptions import UnknownFileTypeError


class File(LoggingMixin):
    """
    The base class for any type of file we want to sign.
    """

    SCHEMA_VERSION = 1
    SUPPORTED_TYPES = ()

    def __init__(self, source, public_key_cache):
        """
        :param source: Typically this will be a file path, but some modules may
               support other types as well.  See ``images.JpegFile`` for an
               example.
        :param public_key_cache: The location on-disk where you're caching
               public keys that have been downloaded for use in verification.
        """

        if isinstance(source, six.string_types):
            if not os.path.exists(source):
                raise FileNotFoundError(
                    "Can't find the file you want to sign/verify: {}".format(
                        source
                    )
                )

        self.source = source
        self.public_key_cache = public_key_cache

    @classmethod
    def build(cls, path, public_key_cache):
        """
        Attempt to find a subclass of File that understands this file and
        return an instance of that class.

        :param path: A path to a file we want to sign/verify.
        :param public_key_cache: The location of the local public key cache.
        :return: An instance of the relevant File subclass
        """

        mimetype = magic.from_file(path, mime=True)
        for klass in cls.get_subclasses():
            if mimetype in klass.SUPPORTED_TYPES:
                return klass(path, public_key_cache)
        raise UnknownFileTypeError()

    @classmethod
    def get_subclasses(cls):
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            if hasattr(subclass, "SUPPORTED_TYPES"):
                yield subclass

    def get_raw_data(self):
        """
        This should return the raw binary data of file -- the part that
        contains the media, so no header data.  This is accomplished in a
        variety of ways, depending on the file format.
        """
        raise NotImplementedError()

    def sign(self, private_key, public_key_url):
        """
        Override this method to generate a signature from the raw data of your
        particular file format and write it to the metadata layer in the
        following format:

          {"version": int, "public-key": url, "signature": signature}

        :param private_key     key  The private key used for signing
        :param public_key_url  str  The URL where you're storing the public key

        :return None
        """
        raise NotImplementedError()

    def verify(self):
        """
        Attempt to verify the origin of a file by checking its local signature
        against the public key listed in the file.
        :return: boolean  ``True`` if verified, `False`` if not.
        """
        raise NotImplementedError()

    def generate_signature(self, private_key):
        """
        Use the private key to generate a signature from raw image data.

        :param private_key: The private key with which we sign the data.
        :return: str A signature, encoded as base64
        """
        return binascii.hexlify(private_key.sign(
            self.get_raw_data(),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        ))

    def generate_payload(self, public_key_url, signature):
        return json.dumps({
            "version": self.SCHEMA_VERSION,
            "public-key": public_key_url,
            "signature": signature.decode()
        }, separators=(",", ":"))

    def verify_signature(self, key_url, signature):
        """
        Use the public key (found either by fetching it online or pulling it
        from the local cache to verify the signature against the image data.
        This method raises an exception on failure, returns None on a pass.

        :param key_url: The URL for the public key we'll use to verify the file
        :param signature: The signature found in the file
        :return: None
        """
        try:
            self._get_public_key(key_url).verify(
                binascii.unhexlify(signature),
                self.get_raw_data(),
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH
                ),
                hashes.SHA256()
            )
        except InvalidSignature:
            self.logger.error("Bad signature")
            return False

        return True

    def _get_public_key(self, url):
        """
        Attempt to fetch the public key from the local cache, and if it's not
        in there, fetch it from the internetz and put it in there.
        :param url: The URL for the public key's location
        :return: The public key
        """

        os.makedirs(self.public_key_cache, exist_ok=True)

        cache = os.path.join(
            self.public_key_cache,
            hashlib.sha512(url.encode("utf-8")).hexdigest()
        )

        if os.path.exists(cache):
            with open(cache, "rb") as f:
                return serialization.load_pem_public_key(
                     f.read(),
                     backend=default_backend()
                )

        response = requests.get(url)
        if response.status_code == 200:
            if b"BEGIN PUBLIC KEY" in response.content:
                with open(cache, "wb") as f:
                    f.write(requests.get(url).content)
                return self._get_public_key(url)

        raise RuntimeError("The specified URL does not contain a public key")


class LargeFile(File):
    """
    For larger files like audio & video, the signature methods are a little
    different so we don't end up busting our RAM limits.
    """

    def generate_signature(self, private_key):

        block_size = 16 * 1024
        raw_data = self.get_raw_data()

        chosen_hash = hashes.SHA512()
        hasher = hashes.Hash(chosen_hash, default_backend())

        buffer = raw_data.read(block_size)
        while len(buffer) > 0:
            hasher.update(buffer)
            buffer = raw_data.read(block_size)

        return binascii.hexlify(private_key.sign(
            hasher.finalize(),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA512()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            utils.Prehashed(chosen_hash)
        ))

    def verify_signature(self, key_url, signature):

        block_size = 16 * 1024

        chosen_hash = hashes.SHA512()
        hasher = hashes.Hash(chosen_hash, default_backend())
        raw = self.get_raw_data()

        buffer = raw.read(block_size)
        while len(buffer) > 0:
            hasher.update(buffer)
            buffer = raw.read(block_size)

        try:
            self._get_public_key(key_url).verify(
                binascii.unhexlify(signature),
                hasher.finalize(),
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA512()),
                    salt_length=padding.PSS.MAX_LENGTH
                ),
                utils.Prehashed(chosen_hash)
            )
        except InvalidSignature:
            self.logger.error("Bad signature")
            return False

        return True



import os
import logging
import shutil

from hashlib import sha512
from unittest.mock import patch

from aletheia.aletheia import Aletheia
from .base import TestCase


class StackTestCase(TestCase):

    TEST_FILES = {
      "jpg": os.path.normpath(
        os.path.join(os.path.dirname(__file__), "data", "test.jpg")),
      "mp3": os.path.normpath(
        os.path.join(os.path.dirname(__file__), "data", "test.mp3")),
    }

    def __init__(self, *args):
        super(StackTestCase, self).__init__(*args)
        logging.basicConfig(level=logging.DEBUG)

    @patch.dict("os.environ", {"ALETHEIA_HOME": TestCase.SCRATCH})
    def test_stack(self):

        aletheia = Aletheia()

        self.assertEqual(
            aletheia.public_key_path,
            os.path.join(self.SCRATCH, "aletheia.pub")
        )
        self.assertEqual(
            aletheia.private_key_path,
            os.path.join(self.SCRATCH, "aletheia.pem")
        )
        self.assertEqual(
            aletheia.public_key_cache,
            os.path.join(self.SCRATCH, "public-keys")
        )

        # Generate the keys

        self.assertFalse(
            os.path.exists(os.path.join(self.SCRATCH, "aletheia.pem")))
        self.assertFalse(
            os.path.exists(os.path.join(self.SCRATCH, "aletheia.pub")))

        aletheia.generate()

        self.assertTrue(
            os.path.exists(os.path.join(self.SCRATCH, "aletheia.pem")))
        self.assertTrue(
            os.path.exists(os.path.join(self.SCRATCH, "aletheia.pub")))

        for suffix, source_path in self.TEST_FILES.items():

            # Copy our test file to SCRATCH so we can fiddle with it

            file_path = os.path.join(self.SCRATCH, "test.{}".format(suffix))
            shutil.copyfile(source_path, file_path)

            # Sign the file

            public_key_url = "https://example.com/aletheia.pub"
            aletheia.sign(file_path, public_key_url)
            with open(source_path, "rb") as original:
                with open(file_path, "rb") as modified:
                    self.assertNotEqual(
                        sha512(original.read()),
                        sha512(modified.read())
                    )

            # Put the public key in the cache so we don't try to fetch it.

            shutil.copyfile(
                os.path.join(self.SCRATCH, "aletheia.pub"),
                os.path.join(
                    self.SCRATCH,
                    "public-keys",
                    sha512(public_key_url.encode("utf-8")).hexdigest()
                )
            )

            # Verify the file

            self.assertTrue(aletheia.verify(file_path))



import logging
import os
import shutil
from unittest import TestCase as BaseTestCase

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key
)


class TestCase(BaseTestCase):

    DATA = os.path.join(os.path.dirname(__file__), "data")
    SCRATCH = os.getenv("ALETHEIA_SCRATCH", "/tmp/aletheia-tests")

    def __init__(self, *args):
        super(TestCase, self).__init__(*args)
        logging.basicConfig(level=logging.DEBUG)

    def setUp(self):
        shutil.rmtree(self.SCRATCH, ignore_errors=True)
        os.makedirs(os.path.join(self.SCRATCH, "public-keys"), exist_ok=True)

    def tearDown(self):
        shutil.rmtree(self.SCRATCH, ignore_errors=True)

    def get_private_key(self):
        with open(os.path.join(self.DATA, "key.pem"), "rb") as f:
            return load_pem_private_key(f.read(), None, default_backend())

    def get_public_key(self):
        with open(os.path.join(self.DATA, "key.pub"), "rb") as f:
            return load_pem_public_key(f.read(), default_backend())

    def copy_for_work(self, name):
        """
        Copy our test file to SCRATCH so we can fiddle with it.
        """
        path = os.path.join(self.SCRATCH, name)
        shutil.copyfile(os.path.join(self.DATA, name), path)
        return path
